module ListTags where

import Info
import Options (TagSort (..))
import TagTransforms

import Control.Monad
import Data.Foldable
import qualified Data.HashMap.Strict as HashMap
import qualified Data.HashSet as HashSet
import Data.List (sort, sortBy)
import Data.Ord (comparing)
import Data.Tuple (swap)
import Text.Printf (printf)


type TagStats = HashMap Text Int
type PathSet  = HashSet FilePath
type FileWarnings = [(FilePath, TagWarning)]

data Stats =
  Stats {
    tags     :: TagStats,
    untagged :: PathSet,
    warns    :: FileWarnings
  }

instance Semigroup Stats where
  s1 <> s2 = Stats {
    tags     = HashMap.unionWith (+) s1.tags s2.tags,
    untagged = s1.untagged <> s2.untagged,
    warns    = s1.warns <> s2.warns
  }

instance Monoid Stats where
  mempty = Stats {tags = [], untagged = [], warns = []}


stats1 :: TagTransforms -> Bool -> FilePath -> Info -> Stats
stats1 tt nsfw path (Info {tags, nsfwTags}) =
  let startTags        = if nsfw then tags <> nsfwTags else tags
      TR outTags warns = applyTransforms tt startTags
  in
  if null outTags then
    Stats {tags = [], untagged = [path], warns = map (path,) warns}
  else
    let tagMap = HashMap.fromList $ map (, 1) $ HashSet.toList outTags in
    Stats {tags = tagMap, untagged = [], warns = map (path,) warns}

stats :: IndexInfo -> Bool -> [(FilePath, Info)] -> Stats
stats iinfo nsfw = foldMap $ uncurry $ stats1 iinfo.tags nsfw

run :: IndexInfo
    -> Bool -- ^ include nsfw?
    -> Bool -- ^ list files with no tags
    -> Bool -- ^ print tag warnings
    -> TagSort
    -> [(FilePath, Info)] ->
    IO ()
run iinfo nsfw listUntagged showWarnings ts infos = do
  let Stats {tags, untagged, warns} = stats iinfo nsfw infos
  let sortedTags = sortBy (cmp ts) $ map swap $ HashMap.toList tags
  putStrLn "TAGS\n----"
  printf "%5d: [total]\n" (length infos)
  for_ sortedTags \(count, path) ->
    printf "%5d: %s\n" count path
  when (listUntagged && not (null untagged)) $ do
    let sortedUntagged = sort $ HashSet.toList untagged
    putStrLn "\nUNTAGGED FILES\n--------------"
    for_ sortedUntagged $ printf " - %s\n"
  when (showWarnings && not (null warns)) $ do
    let hdr :: String = printf "WARNINGS (%d)" (length warns)
    printf "\n%s\n%s\n" hdr (const '-' <$> hdr)
    for_ warns \(file, warn) -> putStrLn $ showWarning file warn

cmp :: TagSort -> (Int, Text) -> (Int, Text) -> Ordering
cmp SortFreq = flip $ comparing fst
cmp SortName = comparing snd
